!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! dosemu DOS/V driver by Takeshi Sone (ts1@tsn.or.jp)
!
! This code is based on:
! Pseudo DOS/V Font/Display Driver for Fep Bridge 0.91
!	by Yutaka Tamiya (tamy@flab.fujitsu.co.jp) 1995/08/20
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
.text

.globl _main
_main:		! for ld86

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Device Header
!	entry font driver of '$IBMAFNT'
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	.word	DispDevHeader	! link to the next driver (offset)
	.word	0		! link to the next driver (segment)
	.word	0x8000		! character device
	.word	Strategy	! strategy routine
	.word	Interrupt	! interrupt routine
	.ascii	"$IBMAFNT"	! device name

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Device Header
!	entry display driver of '$IBMADSP'
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
DispDevHeader:
	.word	-1		! link to the next driver (offset)
	.word	-1		! link to the next driver (segment)
	.word	0x8040		! character device, supports generic ioctl
	.word	Strategy	! strategy routine
	.word	Interrupt	! interrupt routine
	.ascii	"$IBMADSP"	! device name

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Work Area
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
PacketPtr:
	.word	0		! offset
	.word	0		! segment
OrigInt15:
	.long	0

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Strategy Routine
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
Strategy:
	seg cs
	mov [PacketPtr], bx	! offset
	seg cs
	mov [PacketPtr+2], es	! segment
	retf

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Int 15h Handler
!	Int 15h / AX=4900h results in:
!	CF=0, AH=0, BL=0, if the system supports DOS/V Bios
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
Int15dummy:
	cmp ax, #0x4900
	jnz chain_orig_int15
	push bp
	mov bp, sp
	andb [bp+6], #0xfe	! clear CF on the stack
	pop bp
	mov ah, #0
	mov bl, #0
	iret
chain_orig_int15:
	seg cs
	jmpf [OrigInt15]

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Interrupt Routine
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
Interrupt:
	pusha
	push ds
	push es

	mov ax, cs
	mov ds, ax

	les bx, [PacketPtr]
	seg es
	cmpb [bx+2], #0		! init request
	jnz not_init

	cmp [OrigInt15], #0
	jnz already_init
	call Init
	jmp exit_ax
already_init:
	seg es
	mov [bx+14], #Init
	seg es
	mov [bx+14+2], cs
	mov ax, #0x100
	jmp exit_ax

not_init:
	seg es
	cmpb [bx+2], #0x13	! generic ioctl call
	jnz not_g_ioctl
	call g_ioctl
	jmp exit_ax

not_g_ioctl:
	mov ax, #0x8103		! unknown command error

exit_ax:
	seg es
        mov [bx+3], ax

	pop es
	pop ds
	popa
	retf

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Generic IOCTL request
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
g_ioctl:
	seg es
	cmpb [bx+0x0e], #0x4a	! function, select code page
	jz  do_g_ioctl
	mov ax, #0x810c		! general failure error
	ret
do_g_ioctl:
	! call dosemu
	push es
	push bx
	seg es
	les bx, [bx+0x13]	! parameter block 
	seg es
	mov dx, [bx+2]		! code page
	seg es
	cmp [bx], #5		! block size
	jc g_no_dbcs
	add bx, #4		! DBCS lead byte table in es:bx
	jmp g_do_dos_helper
g_no_dbcs:
	xor bx, bx
	mov es, bx
g_do_dos_helper:
	mov ax, #0x90
	int 0xe6		! call dos_helper

	! call BILING.SYS
	mov ax, #0x4f03
	cmp bx, #0
	jz call_biling
	seg es
	cmp [bx], #0
	jz call_biling
	mov al, #2
call_biling:
	int 0x2f

	pop bx
	pop es
	mov ax, #0x100		! command done
	ret

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Initialize
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
Init:
	push es
	push bx

	! save original int 15h handler
	mov ax, #0
	mov es, ax
	seg es
	mov ax, [0x15*4]
	mov [OrigInt15], ax
	seg es
	mov ax, [0x15*4+2]
	mov [OrigInt15+2], ax

	! set int 15h dummy handler
	seg es
	mov [0x15*4], #Int15dummy	! entry Int15dummy
	seg es
	mov [0x15*4+2], cs

	! call BILING.SYS of MS-DOS/V
	! no installation check or error check here; it'd be ok anyway
	mov ax, #0x4f02			! DBCS mode
	int 0x2f

	! notify dosemu to enter DOS/V mode
	mov ax, #0x6601
	int 0x21			! Get code page
	mov dx, #437			! Default (US)
	jc no_codepage
	mov dx, bx
no_codepage:
	push ds
	push dx
	mov ax, #0x6300			! Get DBCS lead byte table
	int 0x21
	or al, al
	jnz no_dbcs
	mov ax, ds
	mov bx, si
	jmp do_dos_helper
no_dbcs:
	xor bx, bx
        mov ax, bx
do_dos_helper:
	pop dx
	pop ds
	mov es, ax
	mov ax, #0x0090		! set dosemu DOS/V emulation mode
	int 0xe6		! call dos_helper in dosemu

	! print message
	mov ah, #9
	mov dx, #Message
	int 0x21

	! save the end address
	pop bx
	pop es
	seg es
	mov [bx+14], #Init
	seg es
	mov [bx+14+2], cs

	mov ax, #0x100	! no error status

	ret

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Message String
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
Message:
	.ascii	"emuDISP.SYS installed."
	.byte	0x0d, 0x0a
	.ascii	"$"
